// -*- c++ -*-
/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#ifndef __mico_fixed_h__
#define __mico_fixed_h__

/*
 * This is a much simplified implementation of the IDL type fixed<d,s>.
 * - binary operations require two arguments of the same type T and
 *   the result is of type T as well
 * - the number of exact digits depends on the size of long double.
 *   with the usual 64-bit mantissa you get no more than 16 exact digits,
 *   a 114-bit mantissa allows for up to 31 exact digits.
 */

#include <iostream.h>

class FixedBase {
    MICO_UShort _digits;
    MICO_Short _scale;
    MICO_LongDouble _val;

    FixedBase ();
    void truncate ();
    MICO_LongDouble transform (MICO_Short nscale) const;
public:
    typedef SequenceTmpl<MICO_OctetWrapper> FixedValue;
    typedef TSeqVar<FixedValue> FixedValue_var;

    FixedBase (MICO_UShort d, MICO_Short s);
    FixedBase (const FixedBase &);
    FixedBase &operator= (const FixedBase &);
    ~FixedBase ();

    MICO_LongDouble to_longdouble () const;
    void from_longdouble (MICO_LongDouble val);

    static void compute_params (MICO_LongDouble val,
				MICO_UShort &digits, MICO_Short &scale);
    /*
     * FixedValue contains '_digits' digits (starting with the most
     * significant one) plus one Octet at the end that is non-zero iff
     * the number is negative.
     */
    FixedValue* to_digits () const;
    void from_digits (const FixedValue &);

    istream &read (istream &);
    ostream &write (ostream &) const;

    static void add (FixedBase &res, const FixedBase &a1, const FixedBase a2);
    static void sub (FixedBase &res, const FixedBase &a1, const FixedBase a2);
    static void mul (FixedBase &res, const FixedBase &a1, const FixedBase a2);
    static void div (FixedBase &res, const FixedBase &a1, const FixedBase a2);
    static void neg (FixedBase &res, const FixedBase &a);

    static int eq (const FixedBase &a1, const FixedBase &a2);
    static int lt (const FixedBase &a1, const FixedBase &a2);
    static int le (const FixedBase &a1, const FixedBase &a2);
};


static inline istream&
operator>> (istream &i, FixedBase &f)
{
    return f.read (i);
}

static inline ostream&
operator<< (ostream &o, const FixedBase &f)
{
    return f.write (o);
}


template<MICO_UShort D, MICO_Short S>
class Fixed : public FixedBase {
public:
    Fixed (int val = 0)
        : FixedBase (D, S)
    {
	from_longdouble (val);
    }
    Fixed (MICO_LongDouble val)
        : FixedBase (D, S)
    {
	from_longdouble (val);
    }
    Fixed (const Fixed<D,S> &f)
        : FixedBase (f)
    {
    }
    ~Fixed ()
    {
    }

    operator MICO_LongDouble () const
    {
	return to_longdouble();
    }

    Fixed<D,S> &operator= (const Fixed<D,S> &f)
    {
	(FixedBase &)*this = (const FixedBase &)f;
	return *this;
    }

    Fixed<D,S> operator++ ()
    {
	Fixed<D,S> res;
	add (res, *this, Fixed<D,S> ((MICO_LongDouble)1.0));
	*this = res;
	return *this;
    }

    Fixed<D,S> operator++ (int)
    {
	Fixed<D,S> res = *this;
	add (res, *this, Fixed<D,S> ((MICO_LongDouble)1.0));
	*this = res;
	return res;
    }

    Fixed<D,S> operator-- ()
    {
	Fixed<D,S> res;
	sub (res, *this, Fixed<D,S> ((MICO_LongDouble)1.0));
	*this = res;
	return *this;
    }

    Fixed<D,S> operator-- (int)
    {
	Fixed<D,S> res = *this;
	sub (res, *this, Fixed<D,S> ((MICO_LongDouble)1.0));
	*this = res;
	return res;
    }

    Fixed<D,S> operator+ () const
    {
	return *this;
    }

    Fixed<D,S> operator- () const
    {
	Fixed<D,S> res;
	neg (res, *this);
	return res;
    }

    Fixed<D,S> operator+ (const Fixed<D,S> &f) const
    {
	Fixed<D,S> res;
	add (res, *this, f);
	return res;
    }

    Fixed<D,S> operator- (const Fixed<D,S> &f) const
    {
	Fixed<D,S> res;
	sub (res, *this, f);
	return res;
    }

    Fixed<D,S> operator* (const Fixed<D,S> &f) const
    {
	Fixed<D,S> res;
	mul (res, *this, f);
	return res;
    }

    Fixed<D,S> operator/ (const Fixed<D,S> &f) const
    {
	Fixed<D,S> res;
	div (res, *this, f);
	return res;
    }

    Fixed<D,S>& operator+= (const Fixed<D,S> &f)
    {
	Fixed<D,S> res;
	add (res, *this, f);
	*this = res;
	return *this;
    }

    Fixed<D,S>& operator-= (const Fixed<D,S> &f)
    {
	Fixed<D,S> res;
	sub (res, *this, f);
	*this = res;
	return *this;
    }

    Fixed<D,S>& operator*= (const Fixed<D,S> &f)
    {
	Fixed<D,S> res;
	mul (res, *this, f);
	*this = res;
	return *this;
    }

    Fixed<D,S>& operator/= (const Fixed<D,S> &f)
    {
	Fixed<D,S> res;
	div (res, *this, f);
	*this = res;
	return *this;
    }

    int operator<  (const Fixed<D,S> &f) const
    {
	return lt (*this, f);
    }

    int operator<= (const Fixed<D,S> &f) const
    {
	return le (*this, f);
    }

    int operator>  (const Fixed<D,S> &f) const
    {
	return !le (*this, f);
    }

    int operator>= (const Fixed<D,S> &f) const
    {
	return !lt (*this, f);
    }

    int operator== (const Fixed<D,S> &f) const
    {
	return eq (*this, f);
    }

    int operator!= (const Fixed<D,S> &f) const
    {
	return !eq (*this, f);
    }

    int operator! () const
    {
	return *this == Fixed<D,S>(0);
    }

    MICO_UShort fixed_digits () const
    {
	return D;
    }

    MICO_Short fixed_scale () const
    {
	return S;
    }
};

#endif
